<?php
/**
 * Optimized Client Script Manager Class File
 *
 * @author Vadim Gabriel <vadimg88[at]gmail[dot]com>
 * @link http://www.vadimg.com/
 * @copyright Vadim Gabriel
 * @license http://www.yiiframework.com/license/
 *
 */

/**

Requirements
--------------
Yii 1.1.x or above
If using the googleComplier API feature then the 'googleComplier' extension needs to be installed and enabled. [Download & Install](http://www.yiiframework.com/extension/google-complier-extension/).

Description:
--------------
This extension will allow you to automatically combine all script files and css files into a single (or several) script or css files.
Basically this will reduce the HTTP calls for resources files by merging several resources files into a single (or more) files.

####Script files:
Script files are merged based on their position, If you use the 'CClientScript::POS_HEAD' you will end up with a single file for all the script files you've used on that page.
If you use 'CClientScript::POS_HEAD' and 'CClientScript::POS_END' for example then you'll end up with two files for each page on that request, Since those resources are located in different positions.

Usage:
---------------

Using this extension is as simple as adding the following code to the application configuration under the components array:

~~~
[php]
'clientScript' => array(
						'class' => 'ext.clientScript.clientScript',
						'combineScriptFiles' => true, // By default this is set to false, set this to true if you'd like to combine the script files
						'combineCssFiles' => true, // By default this is set to false, set this to true if you'd like to combine the css files
						'googleCompiler' => true, // if you have the googleCompiler extension installed then you can set this to true to compile the script code into an optimized and shorter code.
						'googleCompilerLevel' => 'SIMPLE_OPTIMIZATIONS' // If you enable the above setting then you can specify here the compilation level that can be one of the following:
																		// 'WHITESPACE_ONLY', 'SIMPLE_OPTIMIZATIONS', 'ADVANCED_OPTIMIZATIONS'
),
~~~

Then you'd use the regular 'registerScriptFile' & 'registerCssFile' methods as normal and the files will be combined.


NOTE:
---------------
This is currently in testing phase so this should not be used on a production environment.

Known Issues:
----------------
There are several private class members in the CClientScript class, So merging script blocks and codes and css blocks is impossible. Moreover when using this class css and js blocks might not work properly due to the fact that there are private class members in the class which cannot be overridden.
I have filed a bug report to see if those can be set as protected instead.

Other issues may also arise when using this extension so you should test this out on a development application and environment prior using it on a production environment.

Reporting Issue:
-----------------
Reporting Issues and comments are welcome. For better issue control and management please report any issue to the google project hosting that is listed in the extension overview tab.

**/

class clientScript extends CClientScript
{
	/**
	 * Used to store a unique key for the combined files
	 */
	const CACHE_KEY_PREFIX = 'Yii.ClientScripts.Cache';
	/**
	 * @var scripts file name
	 * Each script files position has it's own combined file name
	 */
	public $scriptsFileName = array( self::POS_HEAD => 'scriptshead.js', self::POS_BEGIN => 'scriptbegin.js', self::POS_END => 'scriptsend.js' );
	/**
	 * @var combined css stylesheet file name
	 */
	public $cssFileName = 'styles.css';
	/**
	 * @var cacheID, if you want to enable cache then enter the cache
	 * component name here or false if you don't want to enable cache of
	 * the combined data
	 */
	public $cacheID = 'cache';
	/**
	 * @var cachingDuration number of seconds to store the cached data in cache
	 * after that time the cache will be invalid and will be regenerated and cached.
	 */
	public $cachingDuration = 1440;
	/**
	 * @var if you want to use the googleComplier then enable this by setting this property
	 * to true. You need to have the googleCompiler extension installed and enabled for this to work.
	 * Basically this will reduce the size of the js code by using one of the options of compiling.
	 * Please refer to the googleComplier extension documentation for more information.
	 */
	public $googleCompiler = false;
	/**
	 * @var string the google compiler complication level
	 * refer to the googleCompiler extension documentation for more information.
	 */
	public $googleCompilerLevel = null; 
	/**
	 * @var boolean if to combine the script files or not
	 */
	public $combineScriptFiles = false;
	/**
	 * @var boolean if to combine the css files or not
	 */
	public $combineCssFiles = false;
	/**
	 * @var combine script codes as well? NOT IMPLEMENTED
	 */
	public $combineScriptCode = false;
	/**
	 * @var combine CSS codes as well? NOT IMPLEMENTED
	 */
	public $combineCssCode = false;
	/**
	 * This is here due to the fact that those properties are set to private
	 * A bug was submitted issue #1180
	 */
	private $_hasScripts=false;
	private $_packages;
	private $_dependencies;
	private $_baseUrl;
	private $_coreScripts=array();
	private $_css=array();
	private $_metas=array();
	private $_links=array();
	
	/**
	 * Inserts the scripts in the head section.
	 * @param string the output to be inserted with scripts.
	 */
	public function renderHead(&$output)
	{
		$html='';
		foreach($this->_metas as $meta)
		{
			$html.=CHtml::metaTag($meta['content'],null,null,$meta)."\n";
		}
		foreach($this->_links as $link)
		{
			$html.=CHtml::linkTag(null,null,null,null,$link)."\n";
		}
			
		if( $this->combineCssFiles )
		{
			$html .= $this->combineCssFiles();
		}	
		else
		{
			foreach($this->cssFiles as $url=>$media)
			{
				$html.=CHtml::cssFile($url,$media)."\n";
			}
		}	
		
		foreach($this->_css as $css)
		{
			$html.=CHtml::css($css[0],$css[1])."\n";
		}
		
		if($this->enableJavaScript)
		{
			if( $this->combineScriptFiles  )
			{
				$html .= $this->combineScriptFiles(self::POS_HEAD);
			}
			else
			{
				if(isset($this->scriptFiles[self::POS_HEAD]))
				{
					foreach($this->scriptFiles[self::POS_HEAD] as $scriptFile)
						$html.=CHtml::scriptFile($scriptFile)."\n";
				}
			}

			if(isset($this->scripts[self::POS_HEAD]))
			{
					$html.=CHtml::script(implode("\n",$this->scripts[self::POS_HEAD]))."\n";
			}
		}

		if($html!=='')
		{
			$count=0;
			$output=preg_replace('/(<title\b[^>]*>|<\\/head\s*>)/is','<###head###>$1',$output,1,$count);
			if($count)
			{
				$output=str_replace('<###head###>',$html,$output);
			}
			else
			{
				$output=$html.$output;
			}
		}
	}
	
	/**
	 * Inserts the scripts at the beginning of the body section.
	 * @param string the output to be inserted with scripts.
	 */
	public function renderBodyBegin(&$output)
	{
		$html='';
		
		if( $this->combineScriptFiles  )
		{
			$html .= $this->combineScriptFiles( self::POS_BEGIN );
		}
		else
		{
			if(isset($this->scriptFiles[self::POS_BEGIN]))
			{
				foreach($this->scriptFiles[self::POS_BEGIN] as $scriptFile)
					$html.=CHtml::scriptFile($scriptFile)."\n";
			}
		}
		
		if(isset($this->scripts[self::POS_BEGIN]))
			$html.=CHtml::script(implode("\n",$this->scripts[self::POS_BEGIN]))."\n";

		if($html!=='')
		{
			$count=0;
			$output=preg_replace('/(<body\b[^>]*>)/is','$1<###begin###>',$output,1,$count);
			if($count)
				$output=str_replace('<###begin###>',$html,$output);
			else
				$output=$html.$output;
		}
	}
	
	/**
	 * Inserts the scripts at the end of the body section.
	 * @param string the output to be inserted with scripts.
	 */
	public function renderBodyEnd(&$output)
	{
		if(!isset($this->scriptFiles[self::POS_END]) && !isset($this->scripts[self::POS_END])
			&& !isset($this->scripts[self::POS_READY]) && !isset($this->scripts[self::POS_LOAD]))
			return;

		$fullPage=0;
		$output=preg_replace('/(<\\/body\s*>)/is','<###end###>$1',$output,1,$fullPage);
		$html='';
		
		if( $this->combineScriptFiles  )
		{
			$html .= $this->combineScriptFiles( self::POS_END );
		}
		else
		{
			if(isset($this->scriptFiles[self::POS_END]))
			{
				foreach($this->scriptFiles[self::POS_END] as $scriptFile)
					$html.=CHtml::scriptFile($scriptFile)."\n";
			}
		}
		
		$scripts=isset($this->scripts[self::POS_END]) ? $this->scripts[self::POS_END] : array();
		if(isset($this->scripts[self::POS_READY]))
		{
			if($fullPage)
				$scripts[]="jQuery(document).ready(function() {\n".implode("\n",$this->scripts[self::POS_READY])."\n});";
			else
				$scripts[]=implode("\n",$this->scripts[self::POS_READY]);
		}
		if(isset($this->scripts[self::POS_LOAD]))
		{
			if($fullPage)
				$scripts[]="window.onload=function() {\n".implode("\n",$this->scripts[self::POS_LOAD])."\n};";
			else
				$scripts[]=implode("\n",$this->scripts[self::POS_LOAD]);
		}
		if(!empty($scripts))
			$html.=CHtml::script(implode("\n",$scripts))."\n";

		if($fullPage)
			$output=str_replace('<###end###>',$html,$output);
		else
			$output=$output.$html;
	}
	/**
	 * Combine the CSS files, if cached enabled then cache the result so we won't have to do that
	 * Every time
	 */
	protected function combineCssFiles( )
	{
		$finalResult = '';
		$fromCache = false;
		
		if($this->cachingDuration>0 && $this->cacheID!==false && ($cache=Yii::app()->getComponent($this->cacheID))!==null)
		{
			$key=self::CACHE_KEY_PREFIX.'.csfiles.'.$this->cssFileName;
			if(($data=$cache->get($key))!==false)
			{
				$finalResult = $data;
				$fromCache = true;
			}
		}
	
		if( ( !$fromCache && count($this->cssFiles) ) )
		{
			foreach($this->cssFiles as $url=>$media)
			{
				$contents = file_get_contents( $url );
				if( $contents )
				{
					// Grab the file contents and append to the final result
					$finalResult .= "\n\n/*** CSS File Combined From: {$url} ***/\n\n";
					$finalResult .= $contents."\n";
				}
			}
		}
		
		if( !$finalResult )
		{
			return '';
		}
		
		// Create a file in the assets directory with the contents
		$runtimeFile = Yii::app()->assetManager->basePath . '/' . $this->cssFileName;
		
		// create the script file
		file_put_contents($runtimeFile, $finalResult);
		
		// Publish the script file
		$fileLink = Yii::app()->assetManager->publish( $runtimeFile );
		
		if(isset($cache))
		{
			$cache->set($key,$finalResult,$this->cachingDuration);
		}
		
		return CHtml::cssFile($fileLink)."\n";
	}
	/**
	 * Combine script files, we combine them based on their position, each is combined in a separate file
	 * to load the required data in the required location.
	 * @param $type CClientScript the type of script files currently combined
	 */
	protected function combineScriptFiles( $type = self::POS_HEAD )
	{
		$finalResult = '';
		$fromCache = false;
		
		if($this->cachingDuration>0 && $this->cacheID!==false && ($cache=Yii::app()->getComponent($this->cacheID))!==null)
		{
			$key=self::CACHE_KEY_PREFIX.'.scriptsfiles.'.$this->scriptsFileName[ $type ];
			if(($data=$cache->get($key))!==false)
			{
				$finalResult = $data;
				$fromCache = true;
				
				// Create a file in the assets directory with the contents
				$runtimeFile = Yii::app()->assetManager->basePath . '/' . $this->scriptsFileName[ $type ];

				// create the script file
				file_put_contents($runtimeFile, $finalResult);

				// Publish the script file
				$fileLink = Yii::app()->assetManager->publish( $runtimeFile );

				if(isset($cache))
				{
					$cache->set($key,$finalResult,$this->cachingDuration);
				}

				return CHtml::scriptFile($fileLink)."\n";
			}
		}
	
		if( ( !$fromCache && isset($this->scriptFiles[$type]) ) )
		{
			foreach($this->scriptFiles[$type] as $scriptFile)
			{
				$contents = file_get_contents( $scriptFile );
				if( $contents )
				{
					// Grab the file contents and append to the final result
					$finalResult .= "\n\n/*** Script File Combined From: {$scriptFile} ***/\n";
					$finalResult .= $contents."\n";
				}
			}
		}
		
		if( !$finalResult )
		{
			return '';
		}
		
		// Did we enable the googleCompiler?
		if( $this->googleCompiler )
		{
			// Do we have the extension enabled and working?
			if(($compiler=Yii::app()->getComponent('googleCompiler'))!==null)
			{
				$level = $this->googleCompilerLevel;
				$returnedCode = $compiler->getCompiledByCode($finalResult, $level, 'text');
				
				if( $returnedCode )
				{
					$finalResult = $returnedCode;
				}
			}
			else
			{
				throw new CException(Yii::t('extension.clientScript', 'Sorry, But to use the google compiler you need to have the googleCompiler extension installed.'));
			}
		}
		
		// Create a file in the assets directory with the contents
		$runtimeFile = Yii::app()->assetManager->basePath . '/' . $this->scriptsFileName[ $type ];
		
		// create the script file
		file_put_contents($runtimeFile, $finalResult);
		
		// Publish the script file
		$fileLink = Yii::app()->assetManager->publish( $runtimeFile );
		
		if(isset($cache))
		{
			$cache->set($key,$finalResult,$this->cachingDuration);
		}
		
		return CHtml::scriptFile($fileLink)."\n";
	}
	
}